using System;

namespace UnityEngine.Rendering.HighDefinition
{
    /// <summary>
    /// A volume component that holds settings for the Color Adjustments effect.
    /// </summary>
    [Serializable, VolumeComponentMenu("Post-processing/Color Adjustments")]
    public sealed class ColorAdjustments : VolumeComponent, IPostProcessComponent
    {
        /// <summary>
        /// Sets the value that HDRP uses to adjust the overall exposure of the Scene, in EV.
        /// </summary>
        [Tooltip("Sets the value that HDRP uses to adjust the overall exposure of the Scene, in EV.")]
        public FloatParameter postExposure = new FloatParameter(0f);

        /// <summary>
        /// Controls the overall range of the tonal values.
        /// </summary>
        [Tooltip("Controls the overall range of the tonal values.")]
        public ClampedFloatParameter contrast = new ClampedFloatParameter(0f, -100f, 100f);

        /// <summary>
        /// Specifies the color that HDRP tints the render to.
        /// </summary>
        [Tooltip("Specifies the color that HDRP tints the render to.")]
        public ColorParameter colorFilter = new ColorParameter(Color.white, true, false, true);

        /// <summary>
        /// Controls the hue of all colors in the render.
        /// </summary>
        [Tooltip("Controls the hue of all colors in the render.")]
        public ClampedFloatParameter hueShift = new ClampedFloatParameter(0f, -180f, 180f);

        /// <summary>
        /// Controls the intensity of all colors in the render.
        /// </summary>
        [Tooltip("Controls the intensity of all colors in the render.")]
        public ClampedFloatParameter saturation = new ClampedFloatParameter(0f, -100f, 100f);

        /// <summary>
        /// Tells if the effect needs to be rendered or not.
        /// </summary>
        /// <returns><c>true</c> if the effect should be rendered, <c>false</c> otherwise.</returns>
        public bool IsActive()
        {
            return postExposure.value != 0f
                || contrast.value != 0f
                || colorFilter != Color.white
                || hueShift != 0f
                || saturation != 0f;
        }
    }
}
